"""
XBRL Formatter Setup Configuration.
"""

#  __  __    ___              _        _     __ _
#  \ \/ /   | _ )     _ _    (_)    __| |   / _` |   ___
#   >  <    | _ \    | '_|   | |   / _` |   \__, |  / -_)
#  /_/\_\   |___/   _|_|_   _|_|_  \__,_|   |___/   \___|
# _|"""""|_|"""""|_|"""""|_|"""""|_|"""""|_|"""""|_|"""""|
# "`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'"`-0-0-'

project = "eba-xbridge"
eba_taxonomy_version = "3.4"
description = f"XBRL-XML to XBRL-CSV converter for EBA Taxonomy (version {eba_taxonomy_version})"
url = ""
version = "1.0.2rc4"
author = "Antonio Olleros, Jesus Simon, Javier Hernandez"
author_email = "antonio.olleros@meaningfuldata.eu, jesus.simon@meaningfuldata.eu, javier.hernandez@meaningfuldata.eu"
copyright = "2024 MeaningfulData"
